# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2015 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

import numpy as np

from . import nonlinear_fitting
from ..utils.misc import flatten
from ..utils.math import unit_normalize
from .nonlinear_fitting import nonlinear_least_squares_fit


def fit_PTp_data(mineral, fit_params, flags, data, data_covariances=[],
                 mle_tolerances=[], param_tolerance=1.e-5,
                 delta_params=None, bounds=None,
                 max_lm_iterations=50, verbose=True):
    """
    Given a mineral of any type, a list of fit parameters
    and a set of P-T-property points and (optional) uncertainties,
    this function returns a list of optimized parameters
    and their associated covariances, fitted using the
    scipy.optimize.curve_fit routine.

    Parameters
    ----------
    mineral : mineral
        Mineral for which the parameters should be optimized.

    fit_params : list of strings
        List of dictionary keys contained in mineral.params
        corresponding to the variables to be optimized
        during fitting. Initial guesses are taken from the existing
        values for the parameters

    flags : string or list of strings
        Attribute names for the property to be fit for the whole
        dataset or each datum individually (e.g. 'V')

    data : 2D numpy array of observed X-P-T-property values

    data_covariances : 3D numpy array of X-P-T-property covariances (optional)
        If not given, all covariance matrices are chosen
        such that all data points have equal weight,
        with all error in the pressure.

    mle_tolerances : numpy array (optional)
        Tolerances for termination of the maximum likelihood iterations.

    param_tolerance : float (optional)
        Fractional tolerance for termination of the nonlinear optimization.

    delta_params : numpy array (optional)
        Initial values for the change in parameters.

    bounds : 2D numpy array (optional)
        Minimum and maximum bounds for the parameters. The shape must be
        (n_parameters, 2).

    max_lm_iterations : integer (default : 50)
        Maximum number of Levenberg-Marquardt iterations.

    verbose : boolean (default : True)
        Whether to print detailed information about the optimization to screen.

    Returns
    -------
    model : instance of fitted model
        Fitting-related attributes are as follows:
            n_dof : integer
                Degrees of freedom of the system
            data_mle : 2D numpy array
                Maximum likelihood estimates of the observed data points
                on the best-fit curve
            jacobian : 2D numpy array
                d(weighted_residuals)/d(parameter)
            weighted_residuals : numpy array
                Weighted residuals
            weights : numpy array
                1/(data variances normal to the best fit curve)
            WSS : float
                Weighted sum of squares residuals
            popt : numpy array
                Optimized parameters
            pcov : 2D numpy array
                Covariance matrix of optimized parameters
            noise_variance : float
                Estimate of the variance of the data normal to the curve
    """

    class Model(object):
        def __init__(self, mineral, data, data_covariances, flags, fit_params,
                     mle_tolerances, delta_params=None, bounds=None):
            self.m = mineral
            self.data = data
            self.data_covariances = data_covariances
            self.flags = flags
            self.fit_params = fit_params
            self.mle_tolerances = mle_tolerances
            if delta_params is None:
                self.delta_params = self.get_params()*1.e-5 + 1.e-10
            else:
                self.delta_params = delta_params
            self.bounds = bounds

        def set_params(self, param_values):
            i = 0

            if self.bounds is not None:
                param_values = np.clip(param_values,
                                       self.bounds[:, 0],
                                       self.bounds[:, 1])

            for param in self.fit_params:
                if isinstance(self.m.params[param], float):
                    self.m.params[param] = param_values[i]
                    i += 1
                else:
                    for j in range(len(self.m.params[param])):
                        self.m.params[param][j] = param_values[i]
                        i += 1

        def get_params(self):
            params = []
            for i, param in enumerate(self.fit_params):
                params.append(self.m.params[param])
            return np.array(flatten([mineral.params[prm]
                                     for prm in fit_params]))

        def function(self, x, flag):
            P, T, p = x
            self.m.set_state(P, T)
            return np.array([P, T, getattr(self.m, flag)])

        def normal(self, x, flag):
            P, T, p = x

            if flag == 'V':
                self.m.set_state(P, T)
                dPdp = -self.m.K_T/self.m.V
                dpdT = self.m.alpha*self.m.V
            elif flag == 'H':
                self.m.set_state(P, T)
                dPdp = 1./((1.-T*self.m.alpha)*self.m.V)
                dpdT = self.m.molar_heat_capacity_p
            elif flag == 'S':
                self.m.set_state(P, T)
                dPdp = -1./(self.m.alpha*self.m.V)
                dpdT = self.m.molar_heat_capacity_p/T
            elif flag == 'gibbs':
                self.m.set_state(P, T)
                dPdp = 1./self.m.V
                dpdT = -self.m.S
            else:
                dP = 1.e5
                dT = 1.
                dPdp = (2.*dP)/(self.function([P+dP, T, 0.], flag)[2]
                                - self.function([P-dP, T, 0.], flag)[2])
                dpdT = (self.function([P, T+dT, 0.], flag)[2]
                        - self.function([P, T-dT, 0.], flag)[2])/(2.*dT)
            dPdT = -dPdp*dpdT
            n = np.array([-1., dPdT, dPdp])
            return unit_normalize(n)

    # If only one property flag is given, assume it applies to all data
    if type(flags) is str:
        flags = np.array([flags] * len(data[:, 0]))

    if len(flags) != len(data):
        raise Exception(f'The number of flags (n = {len(flags)}) must be equal '
                        f'to the number of data (n = {len(data)}).')

    # Apply mle tolerances if they dont exist
    if mle_tolerances == []:
        mineral.set_state(1.e5, 300.)
        mle_tolerance_factor = 1.e-5
        mle_tolerances = np.empty(len(flags))
        for i, flag in enumerate(flags):
            if flag in ['gibbs', 'enthalpy', 'H', 'helmholtz']:
                mle_tolerances[i] = 1.  # 1 J
            else:
                mle_tolerances[i] = mle_tolerance_factor * getattr(mineral,
                                                                   flag)

    # If covariance matrix is not given, apply unit weighting to all pressures
    # (with zero errors on T and p)
    covariances_defined = True
    if data_covariances == []:
        covariances_defined = False
        data_covariances = np.zeros((len(data[:, 0]),
                                     len(data[0]), len(data[0])))
        for i in range(len(data_covariances)):
            data_covariances[i][0][0] = 1.

    model = Model(mineral=mineral,
                  data=data,
                  data_covariances=data_covariances,
                  flags=flags,
                  fit_params=fit_params,
                  delta_params=delta_params,
                  mle_tolerances=mle_tolerances,
                  bounds=bounds)

    nonlinear_least_squares_fit(model,
                                max_lm_iterations=max_lm_iterations,
                                param_tolerance=param_tolerance,
                                verbose=verbose)

    if verbose is True and covariances_defined is True:
        confidence_interval = 0.9
        d = nonlinear_fitting.extreme_values(model.weighted_residuals,
                                             confidence_interval)
        confidence_bound, indices, probabilities = d
        if indices != []:
            print('The function nonlinear_fitting.extreme_values'
                  '(model.weighted_residuals, confidence_interval) '
                  f'has determined that there are {len(indices):d} data points'
                  ' which have residuals which are not expected at the '
                  f'{confidence_interval*100.:.1f}% confidence level '
                  f'(> {confidence_bound:.1f} s.d. away from the model fit).\n'
                  'Their indices and the probabilities of finding '
                  'such extreme values are:')
            for i, idx in enumerate(indices):
                print(f'[{idx:d}]: {probabilities[i]:.4f} '
                      f'({np.abs(model.weighted_residuals[idx]):.1f} s.d. '
                      'from the model)')
            print('You might consider removing them from your fit, '
                  'or increasing the uncertainties in their '
                  'measured values.\n')

    return model


def fit_PTV_data(mineral, fit_params,
                 data, data_covariances=[],
                 delta_params=None, bounds=None,
                 param_tolerance=1.e-5, max_lm_iterations=50,
                 verbose=True):
    """
    A simple alias for the fit_PTp_data for when all the data is volume data
    """

    return fit_PTp_data(mineral=mineral, flags='V',
                        data=data, data_covariances=data_covariances,
                        fit_params=fit_params, param_tolerance=param_tolerance,
                        delta_params=delta_params, bounds=bounds,
                        max_lm_iterations=max_lm_iterations, verbose=verbose)


def fit_XPTp_data(solution, fit_params, flags, data, data_covariances=[],
                  mle_tolerances=[], param_tolerance=1.e-5,
                  delta_params=None, bounds=None,
                  max_lm_iterations=50, verbose=True):
    """
    Given a symmetric solution, a list of fit parameters
    and a set of P-T-property points and (optional) uncertainties,
    this function returns a list of optimized parameters
    and their associated covariances, fitted using the
    scipy.optimize.curve_fit routine.

    Parameters
    ----------
    solution : Solution
        Solution for which the parameters should be optimized.

    fit_params : list of lists
        List of lists corresponding to the variables to be optimized
        during fitting. Each list is either of length two or three.
        The first item of length-2 lists should be a
        dictionary key contained in one of the endmember
        mineral.params, and the second item should be the index of
        the endmember in the solution (indexing starts from 0).
        The first item of length-3 lists should be one of 'E', 'S' or
        'V' (the excess energies, entropies or volumes in each binary).
        The second two items should be the indices of the pair of
        endmembers bounding the binary, in ascending order
        (indexing starts from 0). Initial guesses are taken from the existing
        values for the parameters.

    flags : string or list of strings
        Attribute names for the property to be fit for the whole
        dataset or each datum individually (e.g. 'V')

    data : 2D numpy array of observed X-P-T-property values

    data_covariances : 3D numpy array of X-P-T-property covariances (optional)
        If not given, all covariance matrices are chosen
        such that all data points have equal weight,
        with all error in the pressure.

    mle_tolerances : numpy array (optional)
        Tolerances for termination of the maximum likelihood iterations.

    param_tolerance : float (optional)
        Fractional tolerance for termination of the nonlinear optimization.

    delta_params : numpy array (optional)
        Initial values for the change in parameters.

    bounds : 2D numpy array (optional)
        Minimum and maximum bounds for the parameters. The shape must be
        (n_parameters, 2).

    max_lm_iterations : integer (default : 50)
        Maximum number of Levenberg-Marquardt iterations.

    verbose : boolean (default : True)
        Whether to print detailed information about the optimization to screen.

    Returns
    -------
    model : instance of fitted model
        Fitting-related attributes are as follows:
            n_dof : integer
                Degrees of freedom of the system
            data_mle : 2D numpy array
                Maximum likelihood estimates of the observed data points
                on the best-fit curve
            jacobian : 2D numpy array
                d(weighted_residuals)/d(parameter)
            weighted_residuals : numpy array
                Weighted residuals
            weights : numpy array
                1/(data variances normal to the best fit curve)
            WSS : float
                Weighted sum of squares residuals
            fit_params : list of lists
                The parameter lists in their original form
            fit_params_strings : list of strings
                The parameters in user-readable string form
            popt : numpy array
                Optimized parameters
            pcov : 2D numpy array
                Covariance matrix of optimized parameters
            noise_variance : float
                Estimate of the variance of the data normal to the curve
    """

    class Model(object):
        def __init__(self, solution, data, data_covariances, flags, fit_params,
                     mle_tolerances, delta_params=None, bounds=None):
            self.m = solution
            self.data = data
            self.data_covariances = data_covariances
            self.flags = flags
            self.fit_params = fit_params
            self.fit_params_strings = []
            for p in fit_params:
                if isinstance(p, list):
                    csv_list_mbrs = ",".join([str(i) for i in p[1:]])
                    self.fit_params_strings.append(f'{p[0]} ({csv_list_mbrs})')
                else:
                    self.fit_params_strings.append(p)

            self.mle_tolerances = mle_tolerances
            if delta_params is None:
                self.delta_params = self.get_params()*1.e-5 + 1.e-10
            else:
                self.delta_params = delta_params
            self.bounds = bounds

        def set_params(self, param_values):
            # fit_params is a list of lists
            # if the list has length 2, the first item should be an integer
            # indicating the endmember number in the solution
            # if the list has length 3, the first two items should be endmember
            # numbers, and the third should be the interaction parameter type
            # (E, S or V).
            i = 0

            if self.bounds is not None:
                param_values = np.clip(param_values,
                                       self.bounds[:, 0],
                                       self.bounds[:, 1])

            for param in self.fit_params:
                value = param_values[i]
                if len(param) == 2:
                    key, imbr = param
                    if isinstance(self.m.endmembers[imbr][0].params[key],
                                  float):
                        self.m.endmembers[imbr][0].params[key] = value
                        i += 1
                    else:
                        n_values = len(self.m.endmembers[imbr][0].params[key])
                        for j in range(n_values):
                            self.m.endmembers[imbr][0].params[key][j] = value
                            i += 1
                elif len(param) == 3:
                    key, imbr, jmbr = param
                    ai = self.m.solution_model.alphas[imbr]
                    aj = self.m.solution_model.alphas[jmbr]
                    if key == 'E':
                        self.m.solution_model.We[imbr, jmbr] = 2.*value/(ai*aj)
                    if key == 'S':
                        self.m.solution_model.Ws[imbr, jmbr] = 2.*value/(ai*aj)
                    if key == 'V':
                        self.m.solution_model.Wv[imbr, jmbr] = 2.*value/(ai*aj)

                    i += 1
                else:
                    raise Exception('param length must be two or three')

        def get_params(self):
            params = []
            for param in self.fit_params:
                if len(param) == 2:
                    key, imbr = param
                    value = self.m.endmembers[imbr][0].params[key]
                    if isinstance(value, float):
                        params.append(value)
                    else:
                        params.extend(list(value))

                elif len(param) == 3:
                    key, imbr, jmbr = param
                    ai = self.m.solution_model.alphas[imbr]
                    aj = self.m.solution_model.alphas[jmbr]
                    if key == 'E':
                        params.append(self.m.solution_model.We[imbr, jmbr]
                                      * (ai * aj)/2.)
                    if key == 'S':
                        params.append(self.m.solution_model.Ws[imbr, jmbr]
                                      * (ai * aj)/2.)
                    if key == 'V':
                        params.append(self.m.solution_model.Wv[imbr, jmbr]
                                      * (ai * aj)/2.)
                else:
                    raise Exception('param length must be two or three')
            return np.array(params)

        def function(self, x, flag):
            self.m.set_composition(x[:self.m.n_endmembers])
            P, T, p = x[self.m.n_endmembers:]
            self.m.set_state(P, T)

            f = np.copy(x)
            f[-1] = getattr(self.m, flag)
            return f

        def normal(self, x, flag):
            self.m.set_composition(x[:self.m.n_endmembers])
            P, T, p = x[self.m.n_endmembers:]

            if flag == 'V':
                self.m.set_state(P, T)
                dPdp = -self.m.K_T/self.m.V
                dpdT = self.m.alpha*self.m.V
            elif flag == 'H':
                self.m.set_state(P, T)
                dPdp = 1./((1.-T*self.m.alpha)*self.m.V)
                dpdT = self.m.molar_heat_capacity_p
            elif flag == 'S':
                self.m.set_state(P, T)
                dPdp = -1./(self.m.alpha*self.m.V)
                dpdT = self.m.molar_heat_capacity_p/T
            elif flag == 'gibbs':
                self.m.set_state(P, T)
                dPdp = 1./self.m.V
                dpdT = -self.m.S
            else:
                dP = 1.e5
                dT = 1.
                xP0 = np.copy(x)
                xP1 = np.copy(x)
                xT0 = np.copy(x)
                xT1 = np.copy(x)
                xP0[-3] = xP1[-3] - dP
                xP1[-3] = xP1[-3] + dP
                xT0[-2] = xP1[-2] - dT
                xT1[-2] = xP1[-2] + dT

                dPdp = (2.*dP)/(self.function(xP1, flag)[2]
                                - self.function(xP0, flag)[2])
                dpdT = (self.function(xT1, flag)[2]
                        - self.function(xT0, flag)[2])/(2.*dT)
            dPdT = -dPdp*dpdT
            n = np.zeros(len(x))
            n[-3:] = np.array([-1., dPdT, dPdp])
            return unit_normalize(n)

    # If only one property flag is given, assume it applies to all data
    if type(flags) is str:
        flags = np.array([flags] * len(data[:, 0]))

    if len(flags) != len(data):
        raise Exception(f'The number of flags (n = {len(flags)}) must be equal '
                        f'to the number of data (n = {len(data)}).')

    # Apply mle tolerances if they dont exist
    if mle_tolerances == []:
        solution.set_state(1.e5, 300.)
        mle_tolerance_factor = 1.e-5
        mle_tolerances = np.empty(len(flags))
        for i, flag in enumerate(flags):
            if flag in ['gibbs', 'enthalpy', 'H', 'helmholtz']:
                mle_tolerances[i] = 1.  # 1 J
            else:
                mle_tolerances[i] = mle_tolerance_factor*getattr(solution,
                                                                 flag)

    # If covariance matrix is not given, apply unit weighting to all pressures
    # (with zero errors on T and property)
    covariances_defined = True
    if data_covariances == []:
        covariances_defined = False
        nX = solution.n_endmembers
        data_covariances = np.zeros((len(data[:, 0]),
                                     len(data[0]), len(data[0])))
        for i in range(len(data_covariances)):
            data_covariances[i][nX][nX] = 1.

    model = Model(solution=solution,
                  data=data,
                  data_covariances=data_covariances,
                  flags=flags,
                  fit_params=fit_params,
                  mle_tolerances=mle_tolerances,
                  delta_params=delta_params,
                  bounds=bounds)

    nonlinear_least_squares_fit(model,
                                max_lm_iterations=max_lm_iterations,
                                param_tolerance=param_tolerance,
                                verbose=verbose)

    if verbose is True and covariances_defined is True:
        confidence_interval = 0.9
        v_extreme = nonlinear_fitting.extreme_values(model.weighted_residuals,
                                                     confidence_interval)
        confidence_bound, indices, probabilities = v_extreme
        if indices != []:
            print('The function nonlinear_fitting.extreme_values'
                  '(model.weighted_residuals, confidence_interval) '
                  f'has determined that there are {len(indices):d} '
                  'data points which have residuals which are not '
                  f'expected at the {confidence_interval*100.:.1f}% '
                  'confidence level '
                  f'(> {confidence_bound:.1f} s.d. away from the model fit).\n'
                  'Their indices and the probabilities of '
                  'finding such extreme values are:')
            for i, idx in enumerate(indices):
                print(f'[{idx:d}]: {probabilities[i]:.4f} '
                      f'({np.abs(model.weighted_residuals[idx]):.1f} s.d. '
                      'from the model)')
            print('You might consider removing them from your fit, '
                  'or increasing the uncertainties in '
                  'their measured values.\n')

    return model



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
from .linear_fitting import weighted_constrained_least_squares


def fit_composition_to_solution(solution,
                                fitted_variables,
                                variable_values, variable_covariances,
                                variable_conversions=None,
                                normalize=True):
    """
    Takes a Solution object and a set of variable names and
    associates values and covariances and finds the molar fractions of the
    solution which provide the best fit (in a least-squares sense)
    to the variable values.

    The fitting applies appropriate non-negativity constraints
    (i.e. no species can have a negative occupancy on a site).

    Parameters
    ----------
    solution : burnman.Solution object
        The solution to use in the fitting procedure.

    fitted_variables : list of strings
        A list of the variables used to find the best-fit molar fractions of
        the solution. These should either be elements such as "Fe",
        site_species such as "Fef_B" which would correspond to a
        species labelled Fef on the second site,
        or user-defined variables which are arithmetic sums of
        elements and/or site_species defined in "variable_conversions".

    variable_values : numpy array
        Numerical values of the fitted variables.
        These should be given as amounts; they do not need to be normalized.

    variable_covariances : 2D numpy array
        Covariance matrix of the variables.

    variable_conversions : dictionary of dictionaries or None
        A dictionary converting any user-defined variables into an
        arithmetic sum of element and site-species amounts. For example,
        {'Mg_equal': {'Mg_A': 1., 'Mg_B': -1.}}, coupled with Mg_equal = 0
        would impose a constraint that the amount of Mg would be equal on
        the first and second site in the solution.

    normalize : boolean (default: True)
        If True, normalizes the optimized molar fractions to sum to unity.

    Returns
    -------
    popt : numpy array
        Optimized molar fractions.

    pcov : 2D numpy array
        Covariance matrix corresponding to the optimized molar fractions.

    res : float
        The weighted residual of the fitting procedure.
    """

    n_vars = len(fitted_variables)
    n_mbrs = len(solution.endmembers)

    solution_variables = solution.elements
    solution_variables.extend(solution.solution_model.site_names)

    solution_matrix = np.hstack((solution.stoichiometric_matrix,
                                 solution.solution_model.endmember_noccupancies))

    n_sol_vars = solution_matrix.shape[1]

    if variable_conversions is not None:
        solution_matrix = np.hstack((solution_matrix,
                                     np.zeros((solution_matrix.shape[0],
                                               len(variable_conversions)))))

        for i, (new_var, conversion_dict) in enumerate(variable_conversions.items()):
            assert (new_var not in solution_variables)
            solution_variables.append(new_var)

            for var in conversion_dict.keys():
                solution_matrix[:, n_sol_vars+i] += solution_matrix[:, solution_variables.index(var)]

    # Now, construct A using the fitted variables
    A = np.zeros((n_vars, solution_matrix.shape[0]))
    for i, var in enumerate(fitted_variables):
        A[i, :] = solution_matrix[:, solution_variables.index(var)]

    b = variable_values
    Cov_b = variable_covariances

    # Define the constraints
    # Ensure that element abundances / site occupancies
    # are exactly equal to zero if the user specifies that
    # they are equal to zero.
    S, S_index = np.unique(A, axis=0, return_index=True)
    S = np.array([s for i, s in enumerate(S)
                  if np.abs(b[S_index[i]]) < 1.e-10
                  and any(np.abs(s) > 1.e-10)])
    equality_constraints = [S, np.zeros(len(S))]

    # Ensure all site occupancies are non-negative
    T = np.array([-t for t in np.unique(solution.solution_model.endmember_occupancies.T, axis=0)
                  if any(np.abs(t) > 1.e-10)])
    inequality_constraints = [T, np.zeros(len(T))]

    popt, pcov, res = weighted_constrained_least_squares(A, b, Cov_b,
                                                         equality_constraints,
                                                         inequality_constraints)

    if normalize:
        sump = sum(popt)
        popt /= sump
        pcov /= sump * sump
        res /= sump

    # Convert the variance-covariance matrix from endmember amounts to
    # endmember proportions
    dpdx = (np.eye(n_mbrs) - popt).T  # = (1. - p[i] if i == j else -p[i])
    pcov = dpdx.dot(pcov).dot(dpdx.T)
    return (popt, pcov, res)


def fit_phase_proportions_to_bulk_composition(phase_compositions,
                                              bulk_composition):
    """
    Performs weighted constrained least squares on a set of phase compositions
    to find the amount of those phases that best-fits a given bulk composition.

    The fitting applies appropriate non-negativity constraints
    (i.e. no phase can have a negative abundance in the bulk).

    Parameters
    ----------
    phase_compositions : 2D numpy array
        The composition of each phase. Can be in weight or mole amounts.

    bulk_composition : numpy array
        The bulk composition of the composite.
        Must be in the same units as the phase compositions.

    Returns
    -------
    popt : numpy array
        Optimized phase amounts.

    pcov : 2D numpy array
        Covariance matrix corresponding to the optimized phase amounts.

    res : float
        The weighted residual of the fitting procedure.
    """

    n_phases = len(phase_compositions[0])
    inequality_constraints = [-np.eye(n_phases), np.zeros(n_phases)]
    popt, pcov, res = weighted_constrained_least_squares(phase_compositions,
                                                         bulk_composition,
                                                         None,
                                                         None,
                                                         inequality_constraints)
    return (popt, pcov, res)



import numpy as np


def pressure_to_pressure(old_calibrant, new_calibrant,
                         pressure, temperature, PT_covariance=None):
    """
    Convert from pressure defined by one calibrated equation of
    state of a material to pressure defined by an independent
    calibration of the same material.


    Arguments
    ---------
    old_calibrant : Calibrant object
        The original calibration used to estimate the pressure
    new_calibrant : Calibrant object
        The new calibration from which the pressure is desired
    pressure : float
        The pressure calculated using the old calibration
    temperature : float
        The temperature of the material

    PT_covariance : 2x2 numpy array [optional]
            The pressure-temperature
            variance-covariance matrix


    Returns
    -------
    pressure : float
        The pressure of the calibrant [Pa]

    PT_covariance : 2x2 numpy array (if PT_covariance is provided)
        The pressure-temperature variance-covariance matrix.
    """

    if PT_covariance is None:
        V = old_calibrant.volume(pressure, temperature)
        P = new_calibrant.pressure(V, temperature)
        return P
    else:
        V, var_VPT = old_calibrant.volume(pressure, temperature, PT_covariance)
        VT_covariance = var_VPT[np.ix_([0, 2], [0, 2])]
        P, var_PVT = new_calibrant.pressure(V, temperature, VT_covariance)
        PT_covariance = var_PVT[np.ix_([0, 2], [0, 2])]
        return P, PT_covariance



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

# This module provides higher level chemistry-related functions.

from __future__ import absolute_import
import numpy as np
from scipy.optimize import fsolve

from .. import constants

# Import common lower level functions for backwards compatibility
from ..utils.chemistry import dictionarize_formula, formula_mass
from ..utils.chemistry import formula_to_string, site_occupancies_to_strings

def fugacity(standard_material, assemblage):
    """
        Parameters
        ----------
        standard_material: burnman.Material object
            set_method and set_state should already have been used
            material must have a formula as a dictionary parameter

        assemblage: burnman.Composite object
            set_method and set_state should already have been used

        Returns
        -------
        fugacity : float
            Value of the fugacity of the component with respect to
            the standard material

    """
    component_formula = standard_material.params['formula']
    chemical_potential = assemblage.chemical_potential([component_formula])[0]

    fugacity = np.exp((chemical_potential - standard_material.gibbs)
                      / (constants.gas_constant * assemblage.temperature))
    return fugacity


def relative_fugacity(component_formula, assemblage, reference_assemblage):
    """
        Parameters
        ----------
        component_formula: dictionary
            Chemical formula for which to compute the relative fugacity.

        assemblage: burnman.Composite object
            set_method and set_state should already have been used.

        reference_assemblage: burnman.Composite object
            set_method and set_state should already have been used.

        Returns
        -------
        relative_fugacity : float
            Value of the fugacity of the component in the assemblage
            with respect to the reference_assemblage.

    """
    chemical_potential = assemblage.chemical_potential([component_formula])[0]
    reference_chemical_potential = reference_assemblage.chemical_potential([component_formula])[0]

    relative_fugacity = np.exp((chemical_potential
                                - reference_chemical_potential)
                               / (constants.gas_constant
                                  * assemblage.temperature))
    return relative_fugacity


def equilibrium_pressure(minerals, stoichiometry, temperature,
                         pressure_initial_guess=1.e5):
    """
    Given a list of minerals, their reaction stoichiometries
    and a temperature of interest, compute the
    equilibrium pressure of the reaction.

    Parameters
    ----------
    minerals : list of minerals
        List of minerals involved in the reaction.

    stoichiometry : list of floats
        Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol]

    temperature : float
        Temperature of interest [K]

    pressure_initial_guess : optional float
        Initial pressure guess [Pa]

    Returns
    -------
    pressure : float
        The equilibrium pressure of the reaction [Pa]
    """
    def eqm(P, T):
        gibbs = 0.
        for i, mineral in enumerate(minerals):
            mineral.set_state(P[0], T)
            gibbs = gibbs + mineral.gibbs * stoichiometry[i]
        return gibbs

    pressure = fsolve(eqm, [pressure_initial_guess], args=(temperature))[0]

    return pressure


def equilibrium_temperature(minerals, stoichiometry, pressure, temperature_initial_guess=1000.):
    """
    Given a list of minerals, their reaction stoichiometries
    and a pressure of interest, compute the
    equilibrium temperature of the reaction.

    Parameters
    ----------
    minerals : list of minerals
        List of minerals involved in the reaction.

    stoichiometry : list of floats
        Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol]

    pressure : float
        Pressure of interest [Pa]

    temperature_initial_guess : optional float
        Initial temperature guess [K]

    Returns
    -------
    temperature : float
        The equilibrium temperature of the reaction [K]
    """
    def eqm(T, P):
        gibbs = 0.
        for i, mineral in enumerate(minerals):
            mineral.set_state(P, T[0])
            gibbs = gibbs + mineral.gibbs * stoichiometry[i]
        return gibbs

    temperature = fsolve(eqm, [temperature_initial_guess], args=(pressure))[0]

    return temperature


def invariant_point(minerals_r1, stoichiometry_r1,
                    minerals_r2, stoichiometry_r2,
                    pressure_temperature_initial_guess=[1.e9, 1000.]):
    """
    Given a list of minerals, their reaction stoichiometries
    and a pressure of interest, compute the
    equilibrium temperature of the reaction.

    Parameters
    ----------
    minerals : list of minerals
        List of minerals involved in the reaction.

    stoichiometry : list of floats
        Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol]

    pressure : float
        Pressure of interest [Pa]

    temperature_initial_guess : optional float
        Initial temperature guess [K]

    Returns
    -------
    temperature : float
        The equilibrium temperature of the reaction [K]
    """
    def eqm(PT):
        P, T = PT
        gibbs_r1 = 0.
        for i, mineral in enumerate(minerals_r1):
            mineral.set_state(P, T)
            gibbs_r1 = gibbs_r1 + mineral.gibbs * stoichiometry_r1[i]
        gibbs_r2 = 0.
        for i, mineral in enumerate(minerals_r2):
            mineral.set_state(P, T)
            gibbs_r2 = gibbs_r2 + mineral.gibbs * stoichiometry_r2[i]
        return [gibbs_r1, gibbs_r2]

    pressure, temperature = fsolve(eqm, pressure_temperature_initial_guess)
    return pressure, temperature


def hugoniot(mineral, P_ref, T_ref, pressures, reference_mineral=None):
    """
    Calculates the temperatures (and volumes) along a Hugoniot
    as a function of pressure according to the Hugoniot equation
    U2-U1 = 0.5*(p2 - p1)(V1 - V2) where U and V are the
    internal energies and volumes (mass or molar) and U = F + TS


    Parameters
    ----------
    mineral : mineral
        Mineral for which the Hugoniot is to be calculated.

    P_ref : float
        Reference pressure [Pa]

    T_ref : float
        Reference temperature [K]

    pressures : numpy array of floats
        Set of pressures [Pa] for which the Hugoniot temperature
        and volume should be calculated

    reference_mineral : mineral
        Mineral which is stable at the reference conditions
        Provides an alternative U_0 and V_0 when the reference
        mineral transforms to the mineral of interest at some
        (unspecified) pressure.

    Returns
    -------
    temperatures : numpy array of floats
        The Hugoniot temperatures at pressure

    volumes : numpy array of floats
        The Hugoniot volumes at pressure
    """

    def Ediff(T, mineral, P, P_ref, U_ref, V_ref):
        mineral.set_state(P, T[0])
        U = mineral.helmholtz + T[0] * mineral.S
        V = mineral.V

        return (U - U_ref) - 0.5 * (P - P_ref) * (V_ref - V)

    if reference_mineral is None:
        reference_mineral = mineral

    reference_mineral.set_state(P_ref, T_ref)
    U_ref = reference_mineral.helmholtz + T_ref * reference_mineral.S
    V_ref = reference_mineral.V

    temperatures = np.empty_like(pressures)
    volumes = np.empty_like(pressures)

    for i, P in enumerate(pressures):
        temperatures[i] = fsolve(
            Ediff, [T_ref], args=(mineral, P, P_ref, U_ref, V_ref))[0]
        volumes[i] = mineral.V

    return temperatures, volumes



from __future__ import absolute_import
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import warnings
import pkgutil
from scipy.optimize import brentq
from scipy.special import binom

from . import equation_of_state as eos
from ..constants import gas_constant
from ..utils.math import bracket


class BroshCalphad(eos.EquationOfState):
    """
    Class for the high pressure CALPHAD equation of state by
    :cite:`Brosh2007`.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        X = [1./(1. - params['a'][i-2]
                 + params['a'][i-2]
                 * np.power(1.
                            + i/(3.*params['a'][i-2])*pressure/params['K_0'],
                            1./float(i)))
             for i in range(2, 6)]
        V_c = params['V_0']*np.sum([params['c'][i-2]*np.power(X[i-2], 3.)
                                    for i in range(2, 6)])

        nu = self._theta(pressure, params)/temperature
        dP = 1000.
        dthetadP = (self._theta(pressure+dP/2., params)
                    - self._theta(pressure-dP/2., params))/dP
        V_qh = (3. * params['n'] * gas_constant
                * np.exp(-nu)/(1. - np.exp(-nu)) * dthetadP)  # eq. 6

        f = np.sqrt(1. + 2.*params['b'][1]
                    * (1. + params['delta'][1])*pressure/params['K_0'])
        dIdP = ((1. + params['delta'][1])/(params['K_0']
                                           * (1. + params['b'][1]))
                * np.exp((1. - f)/params['b'][1]))
        V_th = self._C_T(temperature, params)*dIdP

        # V = dG_c/dP + dG_qh/dP - C_T*(dI_P/dP)
        return V_c + V_qh + V_th

    def pressure(self, temperature, volume, params):
        def _delta_volume(pressure):
            return (self.volume(pressure, temperature, params) - volume)

        try:
            sol = bracket(_delta_volume, 300.e9, 1.e5, ())
        except ValueError:
            raise Exception('Cannot find a pressure, perhaps you are outside '
                            'the range of validity for the equation of state?')

        return brentq(_delta_volume, sol[0], sol[1])

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns the isothermal bulk modulus :math:`K_T` :math:`[Pa]`
        as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        dP = 1000.
        dV = (self.volume(pressure + dP/2., temperature, params)
              - self.volume(pressure - dP/2., temperature, params))

        return -volume*dP/dV

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns the adiabatic bulk modulus of the mineral. :math:`[Pa]`.
        """
        if temperature < 1.e-10:
            return self.isothermal_bulk_modulus(pressure, temperature, volume,
                                                params)
        else:
            return (self.isothermal_bulk_modulus(pressure, temperature,
                                                 volume, params)
                    * self.molar_heat_capacity_p(pressure, temperature,
                                                 volume, params)
                    / self.molar_heat_capacity_v(pressure, temperature,
                                                 volume, params))

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns the shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        return 0.

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy of the mineral. :math:`[J/mol]`
        """

        return (self.gibbs_free_energy(pressure, temperature, volume, params)
                - pressure * self.volume(pressure, temperature, params)
                + temperature
                * self.entropy(pressure, temperature, volume, params))

    def _Cp_1bar(self, temperature, params):
        # first, identify which of the piecewise segments we're in
        i = np.argmax([T > temperature
                       for T in list(zip(*params['gibbs_coefficients']))[0]])

        # select the appropriate coefficients
        coeffs = params['gibbs_coefficients'][i][1]
        Cp = -(coeffs[2]
               + 2.*coeffs[3]/temperature/temperature
               + 6.*coeffs[4]/(temperature*temperature*temperature)
               + 12.*coeffs[5]*np.power(temperature, -4.)
               + 90.*coeffs[6]*np.power(temperature, -10.)
               + 2.*coeffs[7]*temperature
               + 6.*coeffs[8]*temperature*temperature
               + 12.*coeffs[9]*temperature*temperature*temperature
               + 42.*coeffs[10]*np.power(temperature, 6.)
               - 0.25*coeffs[11]/np.sqrt(temperature)
               - coeffs[12]/temperature)
        return Cp

    def _S_1bar(self, temperature, params):
        # first, identify which of the piecewise segments we're in
        i = np.argmax([T > temperature
                       for T in list(zip(*params['gibbs_coefficients']))[0]])

        # select the appropriate coefficients
        coeffs = params['gibbs_coefficients'][i][1]
        S = -(coeffs[1]
              + coeffs[2]*(1. + np.log(temperature))
              - coeffs[3]/temperature/temperature
              - 2.*coeffs[4]/(temperature*temperature*temperature)
              - 3.*coeffs[5]*np.power(temperature, -4.)
              - 9.*coeffs[6]*np.power(temperature, -10.)
              + 2.*coeffs[7]*temperature
              + 3.*coeffs[8]*temperature*temperature
              + 4.*coeffs[9]*temperature*temperature*temperature
              + 7.*coeffs[10]*np.power(temperature, 6.)
              + 0.5*coeffs[11]/np.sqrt(temperature)
              + coeffs[12]/temperature)
        return S

    def _gibbs_1bar(self, temperature, params):
        # first, identify which of the piecewise segments we're in
        i = np.argmax([T > temperature
                       for T in list(zip(*params['gibbs_coefficients']))[0]])

        # select the appropriate coefficients
        coeffs = params['gibbs_coefficients'][i][1]
        gibbs = (coeffs[0]
                 + coeffs[1]*temperature
                 + coeffs[2]*temperature*np.log(temperature)
                 + coeffs[3]/temperature
                 + coeffs[4]/(temperature*temperature)
                 + coeffs[5]/(temperature*temperature*temperature)
                 + coeffs[6]*np.power(temperature, -9.)
                 + coeffs[7]*temperature*temperature
                 + coeffs[8]*temperature*temperature*temperature
                 + coeffs[9]*np.power(temperature, 4.)
                 + coeffs[10]*np.power(temperature, 7.)
                 + coeffs[11]*np.sqrt(temperature)
                 + coeffs[12]*np.log(temperature))
        return gibbs

    def _X(self, pressure, params):
        return [1./(1. - params['a'][n-2] + params['a'][n-2]
                    * np.power(1. + float(n)/(3.*params['a'][n-2])
                               * pressure/params['K_0'], 1./float(n)))
                for n in range(2, 6)]  # eq. A2

    def _Gamma(self, n, an, Xn):
        def d(k, Xn):
            return (np.power(Xn, 3. - float(k)) * float(k)
                    / (float(k) - 3.) if k != 3
                    else -3.*np.log(Xn))  # eq. A9

        return (3.*np.power(an, 1. - float(n)) / float(n)
                * np.sum([binom(n, k)
                          * np.power(an - 1., float(n-k)) * d(k, Xn)
                         for k in range(0, n+1)]))  # eq. A9, CHECKED

    def _theta(self, pressure, params):
        # Theta (for quasiharmonic term)
        ab2 = (1./(3.*params['b'][0] - 1.))
        K0b = params['K_0']/(1. + params['delta'][0])  # eq. B1b
        XT2 = 1./(1. - ab2 + ab2
                  * np.power(1. + 2./(3.*ab2)
                             * pressure/K0b, 0.5))  # eq. 6 b of SE2015

        # eq. B1 (6 of SE2015)
        return (params['theta_0']
                * np.exp(params['grueneisen_0'] / (1. + params['delta'][0])
                         * (self._Gamma(2, ab2, XT2)
                            - self._Gamma(2, ab2, 1.))))

    def _interpolating_function(self, pressure, params):
        f = np.sqrt(1. + 2.*params['b'][1]
                    * (1. + params['delta'][1])*pressure/params['K_0'])

        # eq. D2 (9 of SE2015)
        return (1. / (1. + params['b'][1]) * (params['b'][1] + f)
                * np.exp((1. - f)/params['b'][1]))

    def _gibbs_qh(self, temperature, theta, n):
        return (3. * n * gas_constant * temperature
                * np.log(1. - np.exp(-theta/temperature)))  # eq. 5

    def _S_qh(self, temperature, theta, n):
        nu = theta/temperature
        return (3. * n * gas_constant * (nu / (np.exp(nu) - 1.)
                                         - np.log(1. - np.exp(-nu))))

    def _C_T(self, temperature, params):
        # C, which is the (G_qh(t,p0) - G_sgte(t,p0)) term
        G_SGTE = self._gibbs_1bar(temperature, params)
        G_qh0 = self._gibbs_qh(temperature, params['theta_0'], params['n'])
        if temperature < params['T_0']:
            C_T = (temperature * temperature
                   / (2.*params['T_0']) * params['delta_Cpr'])

        else:
            C_T = ((G_qh0 - G_SGTE) + params['delta_Gr']
                   - (temperature - params['T_0'])*params['delta_Sr']
                   + (temperature - params['T_0']/2.)*params['delta_Cpr'])

        return C_T

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy of the mineral. :math:`[J/mol]`
        """
        # Cold compression term, eq. A8
        X = self._X(pressure, params)
        G_c = (params['K_0']*params['V_0']
               * np.sum([params['c'][n-2]*(self._Gamma(n, params['a'][n-2],
                                                       X[n-2])
                                           - self._Gamma(n, params['a'][n-2],
                                                         1.))
                         for n in range(2, 6)]))

        # G_SGTE
        G_SGTE = self._gibbs_1bar(temperature, params)

        # G_qh
        theta = self._theta(pressure, params)
        G_qh = self._gibbs_qh(temperature, theta, params['n'])
        G_qh0 = self._gibbs_qh(temperature, params['theta_0'], params['n'])

        C_T = self._C_T(temperature, params)
        I_P = self._interpolating_function(pressure, params)
        return G_SGTE + G_c + G_qh - G_qh0 + C_T*(1. - I_P)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy of the mineral. :math:`[J/K/mol]`
        """

        S_SGTE = self._S_1bar(temperature, params)

        # S_qh
        theta = self._theta(pressure, params)
        S_qh = self._S_qh(temperature, theta, params['n'])
        S_qh0 = self._S_qh(temperature, params['theta_0'], params['n'])

        # dCdT, which is the (S_qh(t,p0) - S_sgte(t,p0)) term
        if temperature < params['T_0']:
            dC_TdT = temperature / params['T_0'] * params['delta_Cpr']

        else:
            dC_TdT = (-(S_qh0 - S_SGTE)
                      - params['delta_Sr'] + params['delta_Cpr'])

        I_P = self._interpolating_function(pressure, params)
        return S_SGTE + S_qh - S_qh0 - dC_TdT*(1. - I_P)

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns the molar isobaric heat capacity :math:`[J/K/mol]`.
        For now, this is calculated by numerical differentiation.
        """
        dT = 0.1
        if temperature < dT/2.:
            return 0.
        else:
            dS = (self.entropy(pressure, temperature+dT/2., volume, params)
                  - self.entropy(pressure, temperature-dT/2., volume, params))
            return temperature*dS/dT

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns the volumetric thermal expansivity :math:`[1/K]`.
        For now, this is calculated by numerical differentiation.
        """
        dT = 0.1
        if temperature < dT/2.:
            return 0.
        else:
            dV = (self.volume(pressure, temperature+dT/2., params)
                  - self.volume(pressure, temperature-dT/2., params))
            return dV/dT/volume

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns the grueneisen parameter.
        This is a dependent thermodynamic variable in this equation of state.
        """
        Cv = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        if Cv == 0.:
            return 0.
        else:
            return (self.thermal_expansivity(pressure, temperature,
                                             volume, params)
                    * self.isothermal_bulk_modulus(pressure, temperature,
                                                   volume, params)
                    * self.volume(pressure, temperature, params))/Cv

    def calculate_transformed_parameters(self, params):
        """
        This function calculates the "c" parameters of the :cite:`Brosh2007`
        equation of state.
        """
        Zs = pkgutil.get_data('burnman',
                              'data/input_masses/atomic_numbers.dat')
        Zs = Zs.decode('ascii').split('\n')
        Z = {str(sl[0]): int(sl[1])
             for sl in [line.split() for line
                        in Zs if len(line) > 0 and line[0] != '#']}

        nZs = [(n_at, float(Z[el]))
               for (el, n_at) in params['formula'].items()]

        # eq. A2 at 300 TPa
        X3_300TPa = [np.power(1. - params['a'][i-2]
                              + params['a'][i-2]
                              * np.power((1. + float(i)/(3.*params['a'][i-2])
                                          * 300.e12/params['K_0']),
                                         1./float(i)), -3.)
                     for i in range(2, 6)]

        # eq. A2 at 330 TPa
        X3_330TPa = [np.power(1. - params['a'][i-2]
                              + params['a'][i-2]
                              * np.power((1. + float(i)/(3.*params['a'][i-2])
                                          * 330.e12/params['K_0']),
                                         1./float(i)), -3.)
                     for i in range(2, 6)]

        # eq. A6a, m^3/mol
        V_QSM_300TPa = np.sum([n_at
                               * (0.02713
                                  * np.exp(0.97626*np.log(Zi)
                                           - 0.057848 * np.log(Zi)*np.log(Zi)))
                               for (n_at, Zi) in nZs])*1.e-6

        # eq. A6b, m^3/mol
        V_QSM_330TPa = np.sum([n_at
                               * (0.025692
                                  * np.exp(0.97914*np.log(Zi)
                                           - 0.057741*np.log(Zi)*np.log(Zi)))
                               for (n_at, Zi) in nZs])*1.e-6

        A = np.array([[1., 1., 1., 1.],  # eq A3
                      [0., 6., 8., 9.],  # eq A4
                      X3_300TPa,  # eq A5a
                      X3_330TPa])  # eq A5b

        b = np.array([1., 8., V_QSM_300TPa/params['V_0'],
                      V_QSM_330TPa/params['V_0']])

        # does not quite reproduce the published values of c
        # A.c consistently gives b[2], b[3] ~1% larger than Brosh
        return np.linalg.solve(A, b)

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        params['T_0'] = 298.15
        if 'P_0' not in params:
            params['P_0'] = 1.e5

        if 'a' not in params:
            params['a'] = [(float(i)-1.)/(3.*params['Kprime_0'] - 1.)
                           for i in range(2, 6)]  # eq. A2

        if 'c' not in params:
            params['c'] = self.calculate_transformed_parameters(params)

        # Calculate reference values for gibbs free energy and heat capacity
        nur = params['theta_0']/params['T_0']
        G_qhr = self._gibbs_qh(params['T_0'], params['theta_0'], params['n'])
        S_qhr = self._S_qh(params['T_0'], params['theta_0'], params['n'])
        Cp_qhr = (3. * params['n'] * gas_constant
                  * nur * nur * np.exp(nur) / np.power(np.exp(nur) - 1., 2.))

        G_SGTEr = self._gibbs_1bar(params['T_0'], params)
        S_SGTEr = self._S_1bar(params['T_0'], params)
        Cp_SGTEr = self._Cp_1bar(params['T_0'], params)

        params['delta_Cpr'] = (Cp_SGTEr - Cp_qhr)
        params['delta_Gr'] = (G_SGTEr - G_qhr)
        params['delta_Sr'] = (S_SGTEr - S_qhr)

        # Check that all the required keys are in the dictionary
        expected_keys = ['gibbs_coefficients',
                         'V_0', 'K_0', 'Kprime_0',
                         'theta_0', 'grueneisen_0', 'delta', 'b']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['P_0'] < 0.:
            warnings.warn('Unusual value for P_0', stacklevel=2)
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-3:
            warnings.warn('Unusual value for V_0', stacklevel=2)
        if params['K_0'] < 1.e9 or params['K_0'] > 1.e13:
            warnings.warn('Unusual value for K_0', stacklevel=2)
        if params['Kprime_0'] < 0. or params['Kprime_0'] > 10.:
            warnings.warn('Unusual value for Kprime_0', stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Calibrant database

  - :mod:`~burnman.calibrants.Decker_1971`
"""


from . import Decker_1971
from . import tools



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.

"""
Decker_1971
^^^^^^^^^^^
"""

import numpy as np
from ..eos.birch_murnaghan import birch_murnaghan
from ..eos import MGD2
from ..eos import debye
from ..classes.calibrant import Calibrant


class NaCl_B1(Calibrant):
    """
    The NaCl (B1 structured) pressure standard reported by Decker (1971).

    Note: This calibrant is not exactly the same as that proposed by Decker.
    The cold compression curve has here been approximated by a
    Birch-Murnaghan EoS.

    TODO: Make the calibrant exactly match that published by Decker.
    """
    def __init__(self):

        def _pressure_Decker_NaCl(volume, temperature, params):
            p300 = birch_murnaghan(params['V_0'] / volume, params)
            grueneisen = MGD2._grueneisen_parameter(0.,
                                                    params['V_0'] / volume,
                                                    params)
            Debye_T = params['Debye_0'] * np.exp((params['grueneisen_0']
                                                  - grueneisen)
                                                 / params['q_0'])
            Eqh = debye.thermal_energy(temperature, Debye_T, params['n'])
            EqhR = debye.thermal_energy(params['refT'], Debye_T, params['n'])
            dpqh = (Eqh - EqhR) * (grueneisen / volume)
            pressure = p300 + dpqh
            return pressure

        _params_Decker_NaCl = {'V_0': 2.7015e-05,
                               'K_0': 24.02e9,
                               'Kprime_0': 4.7369,
                               'Debye_0': 279,
                               'grueneisen_0': 1.59,
                               'q_0': 0.93,
                               'n': 2.,
                               'refT': 298.15,
                               'P_0': 0,
                               'Z': 4.}

        Calibrant.__init__(self, _pressure_Decker_NaCl, 'pressure',
                           _params_Decker_NaCl)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import scipy.optimize as opt
from ..utils.math import bracket


class Calibrant(object):
    """
    The base class for a pressure calibrant material.

    Initialization of a Calibrant object requires the following parameters:

    calibrant_function : Function
        A function that takes either pressure, temperature and a params
        object as arguments, returning the volume, or takes volume, temperature
        and a params object, returning the pressure.

    calibrant_function_return_type : 'pressure' or 'volume'
        The return type of the calibrant function.

    params : dictionary
        A dictionary containing the parameters required by the
        calibrant function.
    """
    def __init__(self, calibrant_function,
                 calibrant_function_return_type,
                 params):

        if calibrant_function_return_type == 'pressure':
            self.pressure_function = calibrant_function
            self.volume_function = self._volume_using_pressure_function
        elif calibrant_function_return_type == 'volume':
            self.volume_function = calibrant_function
            self.pressure_function = self._pressure_using_volume_function
        else:
            raise Exception('calibrant function return type must either '
                            'be pressure or volume')

        self.calibrant_function = calibrant_function

        self.params = params

    def _volume_using_pressure_function(self, pressure, temperature, params):
        """
        Helper function to compute volume iteratively by Brent's method using
        a function of the form pressure(volume, temperature).
        """
        def func(x):
            return (self.pressure_function(x, temperature, params) - pressure)
        try:
            sol = bracket(func, params['V_0'], 1.e-2 * params['V_0'])
        except ValueError:
            raise ValueError('Cannot find a volume, perhaps you are outside '
                             'of the range of validity for the equation '
                             'of state?')
        return opt.brentq(func, sol[0], sol[1])

    def _pressure_using_volume_function(self, volume, temperature, params):
        """
        Helper function to compute pressure iteratively by Brent's method using
        a function of the form volume(pressure, temperature).
        """
        def func(x):
            (self.volume_function(x, temperature, params) - volume)
        try:
            sol = bracket(func, 0., 300.e9)
        except ValueError:
            raise ValueError('Cannot find a pressure, perhaps you are outside '
                             'of the range of validity for the equation '
                             'of state?')
        return opt.brentq(func, sol[0], sol[1])

    def pressure(self, volume, temperature, VT_covariance=None):
        """
        Returns the pressure of the calibrant as a function of
        volume, temperature and (optionally) a volume-temperature
        variance-covariance matrix.

        Parameters
        ----------
        volume : float
            The volume of the calibrant [m^3/mol]
        temperature : float
            The temperature of the calibrant [K]
        VT_covariance : 2x2 numpy array [optional]
            The volume-temperature
            variance-covariance matrix

        Returns
        -------
        pressure : float
            The pressure of the calibrant [Pa]

        PVT_covariance : 3x3 numpy array (if PT_covariance is provided)
            The pressure-volume-temperature variance-covariance matrix.
        """
        if VT_covariance is None:
            return self.pressure_function(volume, temperature, self.params)
        else:
            # Here we take the centered differences
            # We could alternatively use thermodynamic properties
            # but these have not yet been implemented.
            dV = volume / 1.e7
            dT = 0.01
            PdV0 = self.pressure_function(volume - dV/2., temperature,
                                          self.params)
            PdV1 = self.pressure_function(volume + dV/2., temperature,
                                          self.params)
            PdT0 = self.pressure_function(volume, temperature - dT/2.,
                                          self.params)
            PdT1 = self.pressure_function(volume, temperature + dT/2.,
                                          self.params)
            pressure = (PdV0 + PdV1 + PdT0 + PdT1)/4.

            gradPVT = np.zeros((2, 3))
            gradPVT[:, 1:] = np.eye(2)
            gradPVT[:, 0] = [(PdV1 - PdV0)/dV, (PdT1 - PdT0)/dT]
            PVT_covariance = gradPVT.T.dot(VT_covariance).dot(gradPVT)
            return pressure, PVT_covariance

    def volume(self, pressure, temperature, PT_covariance=None):
        """
        Returns the volume of the calibrant as a function of
        pressure, temperature and (optionally) a pressure-temperature
        variance-covariance matrix.

        Parameters
        ----------
        pressure : float
            The pressure of the calibrant [Pa]
        temperature : float
            The temperature of the calibrant [K]
        PT_covariance : 2x2 numpy array [optional]
            The pressure-temperature
            variance-covariance matrix

        Returns
        -------
        volume : float
            The volume of the calibrant [m^3/mol]

        VPT_covariance : 3x3 numpy array (if VT_covariance is provided)
            The volume-pressure-temperature variance-covariance matrix.
        """
        if PT_covariance is None:
            return self.volume_function(pressure, temperature, self.params)
        else:
            # Here we take the centered differences
            # We could alternatively use thermodynamic properties
            # but these have not yet been implemented.
            dP = 100.
            dT = 0.01
            VdP0 = self.volume_function(pressure - dP/2., temperature,
                                        self.params)
            VdP1 = self.volume_function(pressure + dP/2., temperature,
                                        self.params)
            VdT0 = self.volume_function(pressure, temperature - dT/2.,
                                        self.params)
            VdT1 = self.volume_function(pressure, temperature + dT/2.,
                                        self.params)
            volume = (VdP0 + VdP1 + VdT0 + VdT1)/4.

            gradVPT = np.zeros((2, 3))
            gradVPT[:, 1:] = np.eye(2)
            gradVPT[:, 0] = [(VdP1 - VdP0)/dP, (VdT1 - VdT0)/dT]
            VPT_covariance = gradVPT.T.dot(PT_covariance).dot(gradVPT)
            return volume, VPT_covariance



from __future__ import absolute_import
import re

versionstuff = dict(
    re.findall("(.+) = '(.+)'\n", open('burnman/version.py').read()))

metadata = dict(name='burnman',
                version=versionstuff['version'],
                description=('A thermoelastic and thermodynamic toolkit '
                             'for Earth and planetary sciences'),
                url='https://geodynamics.github.io/burnman/',
                author='The BurnMan Team',
                author_email='bob.myhill@bristol.ac.uk',
                license='GPL',
                long_description_content_type="text/x-rst",
                long_description=('BurnMan is a Python library for generating '
                                  'thermodynamic and thermoelastic models of '
                                  'materials and planetary interiors.'),
                packages=['burnman',
                          'burnman.utils',
                          'burnman.minerals',
                          'burnman.eos',
                          'burnman.calibrants',
                          'burnman.tools',
                          'burnman.classes',
                          'burnman.optimize'],
                package_data={'burnman': ['data/input_*/*']},
                classifiers=['License :: OSI Approved :: GNU General Public '
                             'License v2 or later (GPLv2+)',
                             'Programming Language :: Python :: 3.7',
                             'Programming Language :: Python :: 3.8',
                             'Programming Language :: Python :: 3.9',
                             'Programming Language :: Python :: 3.10'],
                )

# Try to use setuptools in order to check dependencies.
# if the system does not have setuptools, fall back on
# distutils.
try:
    from setuptools import setup
    metadata['install_requires'] = ['numpy', 'matplotlib', 'scipy', 'sympy']
except ImportError:
    from distutils.core import setup


setup(**metadata)
